import json
import logging
import requests
import time

from bert import remote_utils, constants, binding, utils, remote_callback

DELAY: float = .1
STOP_DAEMON: bool = False
logger = logging.getLogger(__name__)

def handle_signal(sig, frame):
  if sig == 2:
    global STOP_DAEMON
    STOP_DAEMON = True
    import sys; sys.exit(0)

  else:
    logger.info(f'Unhandled Signal[{sig}]')

def setup_service() -> None:
  if constants.SERVICE_NAME == None:
    raise NotImplementedError('Missing ENVVar[SERVICE_NAME]')

  if constants.MAIN_SERVICE_NONCE is None:
    raise NotImplementedError(f'Missing ENVVar[MAIN_SERVICE_NONCE]')

  if constants.MAIN_SERVICE_HOST is None:
    raise NotImplementedError(f'Missing ENVVar[MAIN_SERVICE_HOST]')

  remote_utils.RemoteConfig.Update({'nonce': constants.MAIN_SERVICE_NONCE})
  url: str = f'{constants.MAIN_SERVICE_HOST}/{constants.SERVICE_NAME}.register'
  response = requests.post(url, data=json.dumps({'nonce': constants.MAIN_SERVICE_NONCE}), headers={
    'Accept': 'application/json',
    'Content-Type': 'application/json',
  })
  response.raise_for_status()


def run_service():
  if not constants.DEBUG:
    raise NotImplementedError

  if constants.DEBUG:
    job_chain: typing.List[types.FunctionType] = binding.build_job_chain()
    noop_queue: utils.Queue = utils.Queue(binding.NOOP_SPACE)
    job_queue: utils.Queue = utils.Queue(job_chain[0].work_key)

    while STOP_DAEMON is False:
      try:
        details: typing.Dict[str, typing.Any] = next(noop_queue)
      except StopIteration:
        time.sleep(DELAY)

      else:
        job_queue.put(details)

        for job in job_chain:
          logger.info(f'Running Job[{job.func_space}] as [{job.pipeline_type.value}] for [{job.__name__}]')
          job()

        else:
          tail_queue: utils.Queue = utils.Queue(job.done_key)
          for details in tail_queue:
            remote_callback.submit(constants.SERVICE_NAME, details)




import collections
import flask
import hashlib
import marshmallow
import marshmallow.schema
import typing

from bert import remote_exceptions, constants, \
  utils as bert_utils, \
  datasource as bert_datasource

PWN: typing.TypeVar = typing.TypeVar('PWN')
DEFAULT_CONTENT_TYPE: str = 'application/octet-stream'

class RemoteConfig(collections.namedtuple('RemoteConfig', constants.REMOTE_CONFIG_KEYS)):
  @classmethod
  def Update(cls: PWN, updates: typing.Dict[str, str]) -> PWN:
    conn: bert_datasource.RedisConnection = bert_datasource.RedisConnection.ParseURL(constants.REDIS_URL)
    config: typing.Dict[str, str] = {}
    for key in constants.REMOTE_CONFIG_KEYS:
      config[key] = conn.client.hget(constants.REMOTE_CONFIG_SPACE, key) or None
      if config[key]:
        config[key] = config[key].decode(constants.ENCODING)

    for key in updates.keys():
      if not key in constants.REMOTE_CONFIG_KEYS:
        raise NotImplementedError(f'Key not allowed[{key}]')

      conn.client.hset(constants.REMOTE_CONFIG_SPACE, key, updates[key])
      config[key] = updates[key]

    return cls(**config)

  @classmethod
  def Obtain(cls: PWN) -> PWN:
    conn: bert_datasource.RedisConnection = bert_datasource.RedisConnection.ParseURL(constants.REDIS_URL)
    config: typing.Dict[str, str] = {}
    for key in constants.REMOTE_CONFIG_KEYS:
      config[key] = conn.client.hget(constants.REMOTE_CONFIG_SPACE, key) or None
      if config[key]:
        config[key] = config[key].decode(constants.ENCODING)

    return cls(**config)


class GetData(collections.namedtuple('GetData', ['data'])):
  @classmethod
  def Obtain(cls: PWN) -> PWN:
    return cls(data={key: value for key, value in flask.request.args.items() if value})

class PostData(collections.namedtuple('PostData', ['data'])):
  @classmethod
  def Obtain(cls: PWN, schema: marshmallow.Schema = None) -> PWN:
    if flask.request.headers.get('Content-Type', DEFAULT_CONTENT_TYPE) in [
      'application/x-www-form-urlencoded',
      'application/x-www-form-urlencoded; charset=UTF-8',
      'application/x-www-form-urlencoded; charset=utf-8']:
      data: typing.Dict[str, typing.Any] = {key: value for key, value in flask.request.form.items()}

    elif flask.request.headers.get('Content-Type', DEFAULT_CONTENT_TYPE) in ['application/json']:
      data: typing.Dict[str, typing.Any] = {key: value for key, value in flask.request.json.items()}

    else:
      raise NotImplementedError(flask.request.headers.get('Content-Type', DEFAULT_CONTENT_TYPE))

    if schema is None:
      return cls(data=data)

    result: marshmallow.schema.UnmarshalResult = schema().load(data)
    if result.errors:
      raise remote_exceptions.BadRequest('Unable to parse input', result.errors, 400)

    return cls(data=data)

def build_auth_token(service_name: str, entropy: str) -> str:
  noop_space: str = hashlib.sha1(constants.NOOP.encode(constants.ENCODING)).hexdigest()
  route_auth: str = f'{service_name}{noop_space}{entropy}'
  return hashlib.sha256(route_auth.encode(constants.ENCODING)).hexdigest()




import flask
import hashlib
import marshmallow
import marshmallow.fields
import logging

from bert import constants, binding, utils, exceptions, remote_exceptions, remote_utils, remote_callback

from urllib.parse import urlparse

from werkzeug.exceptions import NotFound
from werkzeug.serving import run_simple
from werkzeug.wsgi import DispatcherMiddleware

logger = logging.getLogger()

APP: flask.Flask = flask.Flask(f'{constants.SERVICE_NAME}-webservice')
APP.secret_key = constants.WWW_SECRET
APP.debug = constants.DEBUG
MIDDLEWARE: DispatcherMiddleware = DispatcherMiddleware(NotFound(), {
  '': APP,
})
MIDDLEWARE.secret_key = constants.WWW_SECRET
MIDDLEWARE.debug = constants.DEBUG

def setup_service() -> flask.Flask:
  if constants.SERVICE_NAME == None:
    raise NotImplementedError('Missing ENVVar[SERVICE_NAME]')

  if constants.SERVICE_HOST == None:
    raise NotImplementedError(f'Missing ENVVar[SERVICE_HOST]')

  job_chain: typing.List[types.FunctionType] = binding.build_job_chain()
  noop_queue: utils.Queue = utils.Queue(binding.NOOP_SPACE)

  entry_point: types.FunctionType = job_chain[0]
  if entry_point.schema is None:
    raise exceptions.SchemaRequired

  # Error Handling
  @APP.errorhandler(remote_exceptions.BadRequest)
  def _handle_bad_request(exception: remote_exceptions.BadRequest) -> flask.Response:
    if isinstance(exception, remote_exceptions.BadRequest):
      return exception.as_response()

    raise exception

  route_name__queue: str = f'/{constants.SERVICE_NAME}.queue'
  @APP.route(route_name__queue, methods=['POST'])
  def entry_point_handler() -> flask.Response:
    config: remote_utils.RemoteConfig = remote_utils.RemoteConfig.Obtain()
    if flask.request.headers.get('Authorization', None) != f'Bearer {config.auth_token}':
      raise remote_exceptions.BadRequest('Unauthorized', {}, 403)

    data: remote_utils.PostData = remote_utils.PostData.Obtain(schema=entry_point.schema)
    noop_queue.put(data.data)
    return flask.Response('{}', content_type='application/json', status=200)

  class ShakeSchema(marshmallow.Schema):
    auth_token: str = marshmallow.fields.String(required=True)
    queue_url: str = marshmallow.fields.String(required=True)
    shake_url: str = marshmallow.fields.String(required=True)
    callback_url: str = marshmallow.fields.String(required=True)

  route_name__shake: str = f'/{constants.SERVICE_NAME}.shake'
  @APP.route(route_name__shake, methods=['POST'])
  def entry_point__shake() -> flask.Response:
    config: remote_utils.RemoteConfig = remote_utils.RemoteConfig.Obtain()
    if flask.request.headers.get('Authorization', None) != f'Bearer {config.nonce}':
      raise remote_exceptions.BadRequest('Unauthorized', {}, 403)

    if flask.request.headers.get('Authorization', 'noop') == 'Bearer noop':
      raise remote_exceptions.BadRequest('Unauthorized', {}, 403)

    data = remote_utils.PostData.Obtain(schema=ShakeSchema)
    service_parts = urlparse(constants.SERVICE_HOST)
    shake_parts = urlparse(data.data['shake_url'])
    if shake_parts.path != route_name__shake or shake_parts.netloc != service_parts.netloc:
      raise NotImplementedError

    queue_parts = urlparse(data.data['queue_url'])
    if queue_parts.path != route_name__queue or shake_parts.netloc != service_parts.netloc:
      raise NotImplementedError

    remote_utils.RemoteConfig.Update({
      'nonce': 'noop',
      'auth_token': data.data['auth_token'],
      'callback_url': data.data['callback_url']
    })

    return flask.Response('{}', content_type='application/json', status=200)

  return APP

def run_service() -> DispatcherMiddleware:
  run_simple('0.0.0.0', 8000, MIDDLEWARE, use_reloader=False)




import flask
import json

from bert import exceptions

class BadRequest(exceptions.BertException):
  def __init__(self, message: str, errors: dict, status_code: int = 500) -> None:
    self.message = message
    self.errors = errors
    self.status_code = status_code
  def as_response(self) -> flask.Response:
    return flask.Response(json.dumps({'message': self.message, 'errors': self.errors}), self.status_code, mimetype='application/json')





#!/usr/env/bin python

import enum
import hashlib
import logging
import os
import typing

from urllib.parse import urlparse

NOOP: str = 'noop'
DNS: str = '8.8.8.8'
ENCODING: str = 'utf-8'
DEBUG: bool = False if os.environ.get('DEBUG', 'true') in ['f', 'false', 'no'] else True
DELAY: int = .1
DATETIME_FORMAT: str = '%Y-%m-%dT%H:%M:%SZ'
WWW_SECRET: str = os.environ.get('WWW_SECRET', 'noop')

MAIN_SERVICE_HOST: str = os.environ.get('MAIN_SERVICE_HOST', None)
MAIN_SERVICE_NONCE: str = os.environ.get('MAIN_SERVICE_NONCE', None)

SERVICE_HOST: str = os.environ.get('SERVICE_HOST', None)
SERVICE_NAME: str = os.environ.get('SERVICE_NAME', None)

REDIS_URL: str = os.environ.get('REDIS_URL', 'http://localhost:6379/4')
if SERVICE_NAME:
  REMOTE_CONFIG_SPACE: str = ''.join([SERVICE_NAME, WWW_SECRET])
  REMOTE_CONFIG_SPACE: str = hashlib.sha256(REMOTE_CONFIG_SPACE.encode(ENCODING)).hexdigest()
  REMOTE_CONFIG_KEYS: str = ['nonce', 'auth_token', 'callback_url']

else:
  REMOTE_CONFIG_SPACE: str = None
  REMOTE_CONFIG_SPACE: str = 'remote-config-space'
  REMOTE_CONFIG_KEYS: str = []


class PipelineType(enum.Enum):
  BOTTLE: str = 'Bottle'
  CONCURRENT: str = 'Concurrent'

logger = logging.getLogger(__name__)
logger.info(f'DEBUG[{DEBUG}]')




#!/usr/env/bin python

import argparse
import functools
import importlib
import logging
import multiprocessing
import os
import signal
import time
import typing
import types


from datetime import datetime

logger = logging.getLogger(__name__)

JOBS: typing.Dict[str, types.FunctionType] = {}
STOP_DAEMON: bool = False

def capture_options() -> typing.Any:
  parser = argparse.ArgumentParser()
  parser.add_argument('-n', '--new-module', default=None, required=False)
  parser.add_argument('-m', '--module-name', default='bert')
  parser.add_argument('-o', '--how', default=False, action='store_true')
  parser.add_argument('-f', '--flush-db', action='store_true', default=False)
  parser.add_argument('-l', '--log-error-only', default=True, action="store_false")
  parser.add_argument('-e', '--restart-job', default=True, action="store_false")
  parser.add_argument('-a', '--max-restart', default=10, type=int)
  parser.add_argument('-w', '--web-service', default=False, action='store_true')
  parser.add_argument('-d', '--web-service-daemon', default=False, action='store_true')
  return parser.parse_args()

def setup(options: argparse.Namespace) -> None:
  from bert import constants, datasource
  redis_connection: datasource.RedisConnection = datasource.RedisConnection.ParseURL(constants.REDIS_URL)

  if options.flush_db:
    import redis
    logger.info(f'Flushing Redis DB[{redis_connection.db}]')
    redis_connection.client.flushdb()

def scan_jobs(options):
  global JOBS
  module = importlib.import_module(f'{options.module_name}.jobs')
  for member_name in dir(module):
    if member_name.startswith('_'):
      continue

    member = getattr(module, member_name)
    if type(member) != types.FunctionType:
      continue

    JOBS[member_name] = member

def validate_jobs(options) -> None:
  for job in options.jobs.split(','):
    if job not in JOBS.keys():
      raise NotImplementedError(f'Job[{job}] not found. Available Jobs[{", ".join(JOBS.keys())}]')

def handle_signal(sig, frame):
  if sig == 2:
    global STOP_DAEMON
    STOP_DAEMON = True
    import sys; sys.exit(0)

  else:
    logger.info(f'Unhandled Signal[{sig}]')

def start_webservice(options: 'Namespace') -> None:
  signal.signal(signal.SIGINT, handle_signal)
  from bert import binding, constants, utils, remote_webservice
  logger.info(f'Starting WebService[{constants.SERVICE_NAME}]. Debug[{constants.DEBUG}]')
  remote_webservice.setup_service()
  remote_webservice.run_service()

def start_daemon(options: 'Namespace') -> None:
  signal.signal(signal.SIGINT, handle_signal)
  DELAY: float = 0.1
  from bert import binding, constants, utils, remote_daemon
  logger.info(f'Starting service[{constants.SERVICE_NAME}] Daemon. Debug[{constants.DEBUG}]')
  remote_daemon.setup_service()
  remote_daemon.run_service()

def start_jobs(options):
  signal.signal(signal.SIGINT, handle_signal)
  from bert import binding, constants
  if constants.DEBUG:
    job_chain: typing.List[types.FunctionType] = binding.build_job_chain()
    for job in job_chain:
      logger.info(f'Running Job[{job.func_space}] as [{job.pipeline_type.value}] for [{job.__name__}]')
      job()

  else:
    job_chain: typing.List[types.FunctionType] = binding.build_job_chain()
    for job in job_chain:
      logger.info(f'Running Job[{job.func_space}] as [{job.pipeline_type.value}] for [{job.__name__}]')
      logger.info(f'Job worker count[{job.workers}]')
      processes: typing.List[multiprocessing.Process] = []

      @functools.wraps(job)
      def _job_runner(*args, **kwargs):
        job_restart_count: int = 0
        while job_restart_count < options.max_restart:
          try:
            job(*args, **kwargs)
          except Exception as err:
            if options.log_error_only:
              logger.exception(err)

            else:
              raise err

          else:
            break

          job_restart_count += 1
        else:
          logger.exception(f'Job[{job}] failed {job_restart_count} times')

      for idx in range(0, job.workers):
        logging.info(f'Spawning Process[{idx}]')
        proc: multiprocessing.Process = multiprocessing.Process(target=_job_runner, args=())
        proc.daemon = True
        proc.start()
        processes.append(proc)

      else:
        while not STOP_DAEMON and any([proc.is_alive() for proc in processes]):
          time.sleep(constants.DELAY)

def start_service(options: argparse.Namespace) -> None:
  setup(options)
  scan_jobs(options)
  options = capture_options()
  setup(options)
  scan_jobs(options)
  # validate_jobs(options)
  if options.web_service:
    start_webservice(options)

  elif options.web_service_daemon:
    start_daemon(options)

  else:
    start_jobs(options)

if __name__ in ['__main__']:
  options = capture_options()
  start_service(options)




import marshmallow
import marshmallow.fields
import json
import requests
import requests.auth
import typing

from bert import remote_manager, remote_utils

PWN: typing.TypeVar = typing.TypeVar('PWN')
class CallbackSchema(marshmallow.Schema):
  callback_url: str = marshmallow.fields.String(required=True)
  callback_token: str = marshmallow.fields.String(required=True)

def submit(service_name: str, details: typing.Dict[str, typing.Any]) -> None:
  config: remote_utils.RemoteConfig = remote_utils.RemoteConfig.Obtain()
  class TokenAuth(requests.auth.AuthBase):
    def __call__(self, request) -> None:
      request.headers['Authorization'] = f'Bearer {config.auth_token}'
      return request

  params: str = json.dumps(details)
  response = requests.post(config.callback_url, data=params, headers={
    'Content-Type': 'application/json',
    'Accept': 'application/json'
  }, auth=TokenAuth())
  response.raise_for_status()

def queue(service_name: str, details: typing.Dict[str, typing.Any]) -> None:
  url: str = remote_manager.SERVICE_CONNECTIONS[service_name]['queue_url']
  token: str = remote_manager.SERVICE_CONNECTIONS[service_name]['auth_token']
  class TokenAuth(requests.auth.AuthBase):
    def __call__(self: PWN, request: 'Request') -> 'Request':
      request.headers['Authorization'] = f'Bearer {token}'
      return request

  params: str = json.dumps(details)
  response = requests.post(url, data=params, headers={
    'Content-Type': 'application/json',
    'Accept': 'application/json',
  }, auth=TokenAuth())
  response.raise_for_status()




import collections
import flask
import hashlib
import marshmallow
import marshmallow.fields
import json
import requests
import requests.auth
import typing

from bert import remote_exceptions, constants, remote_utils

SERVICE_CONNECTIONS: typing.Dict[str, typing.Any] = {}

class RegisterSchema(marshmallow.Schema):
  nonce: str = marshmallow.fields.String(required=True)

def register(
  app: flask.Flask,
  service_name: str,
  service_entropy: str,
  service_queue_url: str,
  service_shake_url: str,
  callback_url: str) -> 'Route Auth':

  if service_name in SERVICE_CONNECTIONS.keys():
    raise NotImplementedError(f'Duplicate ServiceName[{service_name}]')

  SERVICE_CONNECTIONS[service_name] = {
    'auth_token': remote_utils.build_auth_token(service_name, service_entropy),
    'queue_url': service_queue_url,
    'shake_url': service_shake_url,
    'callback_url': callback_url
  }
  @app.route(f'/{service_name}.register', methods=['POST'])
  def register_service() -> flask.Response:
    if flask.request.json:
      result = RegisterSchema().load(data=flask.request.json)

    elif flask.request.form:
      result = RegisterSchema().load(data=flask.request.form)

    else:
      raise remote_exceptions.BadRequest(f'Unable to read data', {}, 400)

    if result.errors:
      raise remote_exceptions.BadRequest(f'Unable to register microservice', result.errors, 400)

    class TokenAuth(requests.auth.AuthBase):
      def __call__(self, request: typing.Any) -> 'Request':
        request.headers['Authorization'] = f'Bearer {result.data["nonce"]}'
        return request

    response = requests.post(SERVICE_CONNECTIONS[service_name]['shake_url'], data=json.dumps(SERVICE_CONNECTIONS[service_name]), headers={
      'Content-Type': 'application/json',
      'Accept': 'application/json',
    }, auth=TokenAuth())
    if not response.status_code in [200]:
      if isinstance(response.content, bytes):
        content: str = response.content.decode(constants.ENCODING)
      else:
        content: str = response.content

      raise remote_exceptions.BadRequest(f'Microservice is Offline', {'errors': [{'content': content, 'status': response.status_code}]}, 422)

    return flask.Response('{}', content_type='application/json', status=200)

  return SERVICE_CONNECTIONS[service_name]['auth_token']

def handle_bad_request(exception: remote_exceptions.BadRequest) -> flask.Response:
  return exception.as_response()




#!/usr/bin/env python

import os
import sys

from bert import factory, utils

if __name__ in ['__main__']:
  sys.path.append(os.getcwd())
  options = factory.capture_options()
  factory.start_service(options)

